@@ -1,5 +1,5 @@
 /* hdparm.c - Command line interface to get/set hard disk parameters */
-/*          - by Mark Lord (C) 1994-2007 -- freely distributable */
+/*          - by Mark Lord (C) 1994-2008 -- freely distributable */
 #include <unistd.h>
 #include <string.h>
 #include <stdlib.h>
@@ -21,10 +21,11 @@
 #include <asm/byteorder.h>
 
 #include "hdparm.h"
+#include "sgio.h"
 
 extern const char *minor_str[];
 
-#define VERSION "v7.6"
+#define VERSION "v8.6"
 
 #ifndef O_DIRECT
 #define O_DIRECT	040000	/* direct disk access, not easily obtained from headers */
@@ -87,9 +88,8 @@ static char security_password[33];
 static int get_powermode  = 0, set_powermode = 0;
 static int set_apmmode = 0, get_apmmode= 0, apmmode = 0;
 static int get_cdromspeed = 0, set_cdromspeed = 0, cdromspeed = 0;
-static int do_IDentity = 0, do_drq_hsm_error = 0;
+static int do_IDentity = 0, drq_hsm_error = 0;
 static int get_unregister = 0, set_unregister = 0, unregister = 0;
-static int	hwif = 0;
 static int	scan_hwif = 0;
 static int	hwif_data = 0;
 static int	hwif_ctrl = 0;
@@ -98,10 +98,25 @@ static int	set_busstate = 0, get_busstate = 0, busstate = 0;
 static int	set_reread_partn = 0, get_reread_partn;
 static int	set_acoustic = 0, get_acoustic = 0, acoustic = 0;
 
+static int   make_bad_sector = 0, make_bad_sector_flagged;
+static __u64 make_bad_sector_addr = ~0ULL;
+
+static int   write_sector = 0;
+static __u64 write_sector_addr = ~0ULL;
+
+static int   read_sector = 0;
+static __u64 read_sector_addr = ~0ULL;
+
+static int   set_max_sectors = 0, set_max_permanent, get_native_max_sectors = 0;
+static __u64 set_max_addr = 0;
+
 static int	get_doreset = 0, set_doreset = 0;
 static int	get_tristate = 0, set_tristate = 0, tristate = 0;
 static int	i_know_what_i_am_doing = 0;
 
+const int timeout_12secs = 12;
+const int timeout_2hrs   = (2 * 60 * 60);
+
 static int open_flags = O_RDWR|O_NONBLOCK;
 
 // Historically, if there was no HDIO_OBSOLETE_IDENTITY, then
@@ -138,12 +153,15 @@ static void on_off (unsigned int value)
 
 static void flush_buffer_cache (int fd)
 {
-	fsync (fd);				/* flush buffers */
+	sync();
+	fsync(fd);				/* flush buffers */
+	fdatasync(fd);				/* flush buffers */
+	sync();
 	if (ioctl(fd, BLKFLSBUF, NULL))		/* do it again, big time */
 		perror("BLKFLSBUF failed");
-	/* await completion */
-	if (do_drive_cmd(fd, NULL) && errno != EINVAL && errno != ENOTTY && errno != ENOIOCTLCMD)
-		perror("HDIO_DRIVE_CMD(null) (wait for flush complete) failed");
+	else
+		do_drive_cmd(fd, NULL);	/* IDE: await completion */
+	sync();
 }
 
 static int seek_to_zero (int fd)
@@ -178,7 +196,7 @@ static int read_big_block (int fd, char *buf)
 static void *prepare_timing_buf (unsigned int len)
 {
 	unsigned int i;
-	unsigned char *buf;
+	__u8 *buf;
 
 	buf = mmap(NULL, len, PROT_READ|PROT_WRITE, MAP_ANONYMOUS|MAP_PRIVATE, -1, 0);
 	if (buf == MAP_FAILED) {
@@ -222,7 +240,7 @@ static void time_cache (int fd)
 	fflush(stdout);
 
 	/* Clear out the device request queues & give them time to complete */
-	sync();
+	flush_buffer_cache(fd);
 	sleep(1);
 
 	/* Now do the timing */
@@ -281,17 +299,20 @@ static int do_blkgetsize (int fd, unsigned long long *blksize64)
 	}
 #endif
 	rc = ioctl(fd, BLKGETSIZE, &blksize32);	// returns sectors
-	if (rc)
+	if (rc) {
+		rc = errno;
 		perror(" BLKGETSIZE failed");
+	}
 	*blksize64 = blksize32;
 	return rc;
 }
 
-static void time_device (int fd)
+static int time_device (int fd)
 {
 	char *buf;
 	double elapsed;
 	struct itimerval e1, e2;
+	int err = 0;
 	unsigned int max_iterations = 1024, total_MB, iterations;
 
 	/*
@@ -300,12 +321,15 @@ static void time_device (int fd)
 	if (do_ctimings || do_timings) {
 		unsigned long long blksize;
 		do_flush = 1;
-		if (0 == do_blkgetsize(fd, &blksize))
+		err = do_blkgetsize(fd, &blksize);
+		if (!err)
 			max_iterations = blksize / (2 * 1024) / TIMING_BUF_MB;
 	}
 	buf = prepare_timing_buf(TIMING_BUF_BYTES);
 	if (!buf)
-		return;
+		err = ENOMEM;
+	if (err)
+		goto quit;
 
 	printf(" Timing %s disk reads:  ", (open_flags & O_DIRECT) ? "O_DIRECT" : "buffered");
 	fflush(stdout);
@@ -337,7 +361,9 @@ static void time_device (int fd)
 			total_MB, elapsed, total_MB / elapsed * 1024);
 quit:
 	munlockall();
-	munmap(buf, TIMING_BUF_BYTES);
+	if (buf)
+		munmap(buf, TIMING_BUF_BYTES);
+	return err;
 }
 
 static void dmpstr (const char *prefix, unsigned int i, const char *s[], unsigned int maxi)
@@ -348,6 +374,20 @@ static void dmpstr (const char *prefix, unsigned int i, const char *s[], unsigne
 		printf("%s%s", prefix, s[i]);
 }
 
+static __u64 get_lba_capacity (__u16 *idw)
+{
+	__u64 nsects = (idw[58] << 16) | idw[57];
+
+	if (idw[49] & 0x200) {
+		nsects = (idw[61] << 16) | idw[60];
+		if ((idw[83] & 0xc000) == 0x4000 && (idw[86] & 0x0400)) {
+			nsects = (idw[103] << 16) | idw[102];
+			nsects = (nsects << 32) | ((idw[101] << 16) | idw[100]);
+		}
+	}
+	return nsects;
+}
+
 static void dump_identity (__u16 *idw)
 {
 	int i;
@@ -387,7 +427,7 @@ static void dump_identity (__u16 *idw)
 	printf(" CurCHS=%u/%u/%u, CurSects=%u", idw[54], idw[55], idw[56], idw[57] | (idw[58] << 16));
 	printf(", LBA=%s", YN(idw[49] & 0x200));
 	if (idw[49] & 0x200)
- 		printf(", LBAsects=%u", idw[60] | (idw[61] << 16));
+ 		printf(", LBAsects=%llu", get_lba_capacity(idw));
 
 	if (idw[49] & 0x100) {
 		if (idw[62] | idw[63]) {
@@ -434,12 +474,8 @@ static void dump_identity (__u16 *idw)
 			if (idw[88] & 0x010)	strcat(umodes,"udma4 ");
 			if (idw[88] & 0x2000)	strcat(umodes,"*");
 			if (idw[88] & 0x020)	strcat(umodes,"udma5 ");
-			if (idw[93] & 0x2000) {
-				if (idw[88] & 0x4000)	strcat(umodes,"*");
-				if (idw[88] & 0x0040)	strcat(umodes,"udma6 ");
-				if (idw[88] & 0x8000)	strcat(umodes,"*");
-				if (idw[88] & 0x0080)	strcat(umodes,"udma7 ");
-			}
+			if (idw[88] & 0x4000)	strcat(umodes,"*");
+			if (idw[88] & 0x040)	strcat(umodes,"udma6 ");
 		}
 	}
 	if ((idw[49] & 0x100) && (idw[53] & 2))
@@ -640,7 +676,7 @@ do_set_security (int fd)
 	int err = 0;
 	const char *description;
 	struct hdio_taskfile *r;
-	const unsigned int two_hours = 7200, ten_seconds = 10;
+	__u8 *data;
 
 	r = malloc(sizeof(struct hdio_taskfile) + 512);
 	if (!r) {
@@ -649,22 +685,24 @@ do_set_security (int fd)
 		exit(err);
 	}
 
-	memset(r, 0, sizeof(r));
+	memset(r, 0, sizeof(struct hdio_taskfile) + 512);
 	r->cmd_req	= TASKFILE_CMD_REQ_OUT;
-	r->xfer_method	= TASKFILE_XFER_METHOD_PIO_OUT;
-	r->out_bytes	= 512;
+	r->dphase	= TASKFILE_DPHASE_PIO_OUT;
+	r->obytes	= 512;
 	r->lob.command	= security_command;
-	r->data[0]	= security_master & 0x01;
-	memcpy(&r->data[2], security_password, 32);
+	data		= (__u8*)r->data;
+	data[0]		= security_master & 0x01;
+	memcpy(data+2, security_password, 32);
 
-	/* Not setting any out_flags causes a segfault and most
+	/* Not setting any oflags causes a segfault and most
 	   of the times a kernel panic */
-	r->out_flags.lob.command = 1;
+	r->oflags.b.command = 1;
+	r->oflags.b.feat    = 1;
 
 	switch (security_command) {
 		case ATA_OP_SECURITY_ERASE_UNIT:
 			description = "SECURITY_ERASE";
-			r->data[0] |= (enhanced_erase & 0x02);
+			data[0] |= (enhanced_erase & 0x02);
 			break;
 		case ATA_OP_SECURITY_DISABLE:
 			description = "SECURITY_DISABLE";
@@ -674,11 +712,11 @@ do_set_security (int fd)
 			break;
 		case ATA_OP_SECURITY_SET_PASS:
 			description = "SECURITY_SET_PASS";
-			r->data[1] = (security_mode & 0x01);
+			data[1] = (security_mode & 0x01);
 			if (security_master) {
 				/* master password revision code */
-				r->data[34] = 0x11;
-				r->data[35] = 0xff;
+				data[34] = 0x11;
+				data[35] = 0xff;
 			}
 			break;
 		default:
@@ -686,9 +724,9 @@ do_set_security (int fd)
 			exit(EINVAL);
 	}
 	printf(" Issuing %s command, password=\"%s\", user=%s",
-		description, security_password, r->data[0] ? "master" : "user");
+		description, security_password, data[0] ? "master" : "user");
 	if (security_command == ATA_OP_SECURITY_SET_PASS)
-		printf(", mode=%s", r->data[1] ? "max" : "high");
+		printf(", mode=%s", data[1] ? "max" : "high");
 	printf("\n");
 
 	/*
@@ -710,7 +748,7 @@ do_set_security (int fd)
 			err = errno;
 			perror("ERASE_PREPARE");
 		} else {
-			if ((do_taskfile_cmd(fd, r, two_hours))) {
+			if ((do_taskfile_cmd(fd, r, timeout_2hrs))) {
 				err = errno;
 				perror("SECURITY_ERASE");
 			}
@@ -718,24 +756,24 @@ do_set_security (int fd)
 	} else if (security_command == ATA_OP_SECURITY_DISABLE) {
 		/* First attempt an unlock  */
 		r->lob.command = ATA_OP_SECURITY_UNLOCK;
-		if ((do_taskfile_cmd(fd, r, ten_seconds))) {
+		if ((do_taskfile_cmd(fd, r, timeout_12secs))) {
 			err = errno;
 			perror("SECURITY_UNLOCK");
 		} else {
 			/* Then the security disable */
 			r->lob.command = security_command;
-			if ((do_taskfile_cmd(fd, r, ten_seconds))) {
+			if ((do_taskfile_cmd(fd, r, timeout_12secs))) {
 				err = errno;
 				perror("SECURITY_DISABLE");
 			}
 		}
 	} else if (security_command == ATA_OP_SECURITY_UNLOCK) {
-		if ((do_taskfile_cmd(fd, r, ten_seconds))) {
+		if ((do_taskfile_cmd(fd, r, timeout_12secs))) {
 			err = errno;
 			perror("SECURITY_UNLOCK");
 		}
 	} else if (security_command == ATA_OP_SECURITY_SET_PASS) {
-		if ((do_taskfile_cmd(fd, r, ten_seconds))) {
+		if ((do_taskfile_cmd(fd, r, timeout_12secs))) {
 			err = errno;
 			perror("SECURITY_SET_PASS");
 		}
@@ -783,50 +821,298 @@ static void confirm_i_know_what_i_am_doing (const char *opt, const char *explana
 {
 	if (!i_know_what_i_am_doing) {
 		fprintf(stderr, "Use of %s is VERY DANGEROUS.\n%s\n"
-		"Please supply the --yes-i-know-what-i-am-doing flag if you really want this\n"
-		"Program aborted\n", opt, explanation);
+		"Please supply the --yes-i-know-what-i-am-doing flag if you really want this.\n"
+		"Program aborted.\n", opt, explanation);
 		exit(EPERM);
 	}
 }
 
-static void flush_wcache (int fd, __u16 **id_p)
+static int flush_wcache (int fd, __u16 **id_p)
 {
 	__u8 args[4] = {ATA_OP_FLUSHCACHE,0,0,0};
 	__u16 *id;
+	int err = 0;
 
 	*id_p = id = get_identify_data(fd, *id_p);
 	if (id && (id[83] & 0xe000) == 0x6000)
 		args[0] = ATA_OP_FLUSHCACHE_EXT;
-	if (do_drive_cmd(fd, args))
+	if (do_drive_cmd(fd, args)) {
+		err = errno;
 		perror (" HDIO_DRIVE_CMD(flushcache) failed");
+	}
+	return err;
+}
+
+static void dump_sector (__u16 *w)
+{
+	unsigned int i;
+
+	for (i = 0; i < (256/8); ++i) {
+		printf("%04x %04x %04x %04x %04x %04x %04x %04x\n", w[0], w[1], w[2], w[3], w[4], w[5], w[6], w[7]);
+		w += 8;
+	}
+}
+
+static __u64 get_start_lba (int fd)
+{
+	__u64 lba;
+
+	static struct local_hd_geometry      g;
+	static struct local_hd_big_geometry bg;
+
+	if (!ioctl(fd, HDIO_GETGEO_BIG, &bg)) {
+		lba = bg.start;
+	} else if (!ioctl(fd, HDIO_GETGEO, &g)) {
+		lba = g.start;
+	} else {
+		int err = errno;
+		perror(" HDIO_GETGEO failed: cannot determine correct LBA offset, aborting.");
+		exit(err);
+	}
+	return lba;
+}
+
+static int abort_if_not_full_device (int fd, __u64 lba, const char *devname)
+{
+	__u64 start = get_start_lba(fd);
+
+	char *fdevname = strdup(devname);
+	int i, shortened = 0;
+
+	for (i = strlen(fdevname); --i > 2 && (fdevname[i] >= '0' && fdevname[i] <= '9');) {
+		fdevname[i] = '\0';
+		shortened = 1;
+	}
+
+	if (!shortened)
+		fdevname = "the full disk";
+
+	if (start == 0ULL)
+		return 0;
+	fprintf(stderr, "Device %s has non-zero LBA starting offset of %llu.\n", devname, start);
+	fprintf(stderr, "Please use an absolute LBA with the /dev/ entry for the full device, rather than a partition name.\n");
+	fprintf(stderr, "%s is probably a partition of %s (?)\n", devname, fdevname);
+	fprintf(stderr, "The absolute LBA of sector %llu from %s should be %llu\n", lba, devname, start + lba);
+	fprintf(stderr, "Aborting.\n");
+	exit(EINVAL);
+}
+
+static __u64 do_get_native_max_sectors (int fd, __u16 *id)
+{
+	int err = 0;
+	__u64 max = 0;
+	struct hdio_taskfile r;
+
+	memset(&r, 0, sizeof(r));
+	r.cmd_req = TASKFILE_CMD_REQ_NODATA;
+	r.dphase  = TASKFILE_DPHASE_NONE;
+	r.oflags.b.dev      = 1;
+	r.oflags.b.command  = 1;
+	r.iflags.b.command  = 1;
+	r.iflags.b.lbal     = 1;
+	r.iflags.b.lbam     = 1;
+	r.iflags.b.lbah     = 1;
+	r.lob.dev = 0x40;
+
+	if (((id[83] & 0xc400) == 0x4400) && (id[86] & 0x0400)) {
+		r.iflags.b.hob_lbal  = 1;
+		r.iflags.b.hob_lbam  = 1;
+		r.iflags.b.hob_lbah  = 1;
+		r.lob.command = ATA_OP_READ_NATIVE_MAX_EXT;
+		if (do_taskfile_cmd(fd, &r, 10)) {
+			err = errno;
+			perror (" READ_NATIVE_MAX_ADDRESS_EXT failed");
+		} else {
+			max = (((__u64)((r.hob.lbah << 16) | (r.hob.lbam << 8) | r.hob.lbal) << 24)
+				     | ((r.lob.lbah << 16) | (r.lob.lbam << 8) | r.lob.lbal)) + 1;
+		}
+	} else {
+		r.iflags.b.dev = 1;
+		r.lob.command = ATA_OP_READ_NATIVE_MAX;
+		if (do_taskfile_cmd(fd, &r, timeout_12secs)) {
+			err = errno;
+			perror (" READ_NATIVE_MAX_ADDRESS failed");
+		} else {
+			max = ((r.lob.lbah << 16) | (r.lob.lbam << 8) | r.lob.lbal) + 1;
+		}
+	}
+	errno = err;
+	return max;
+}
+
+static int do_make_bad_sector (int fd, __u16 *id, __u64 lba, const char *devname)
+{
+	int err = 0, has_write_unc;
+	struct hdio_taskfile *r;
+	const char *flagged;
+
+	abort_if_not_full_device(fd, lba, devname);
+	r = malloc(sizeof(struct hdio_taskfile) + 520);
+	if (!r) {
+		err = errno;
+		perror("malloc()");
+		return err;
+	}
+
+	has_write_unc = (id[ 83] & 0xc000) == 0x4000 && (id[ 86] & 0x8000) == 0x8000
+		     && (id[119] & 0xc004) == 0x4004 && (id[120] & 0xc000) == 0x4000;
+
+	if (has_write_unc || make_bad_sector_flagged || lba >> 28) {
+		if (!has_write_unc) {
+			printf("Device does not claim to implement the required WRITE_UNC_EXT command\n"
+				"This operation will probably fail (continuing regardless).\n");
+		}
+		init_hdio_taskfile(r, ATA_OP_WRITE_UNC_EXT, RW_READ, LBA48_FORCE, lba, 1, 0);
+		r->oflags.b.feat = 1;
+		r->lob.feat = make_bad_sector_flagged ? 0xaa : 0x55;
+		flagged     = make_bad_sector_flagged ? "flagged" : "pseudo";
+		printf("Corrupting sector %llu (WRITE_UNC_EXT as %s): ", lba, flagged);
+	} else {
+		init_hdio_taskfile(r, ATA_OP_WRITE_LONG_ONCE, RW_WRITE, LBA28_OK, lba, 1, 520);
+		memset(r->data, 0xff, 520);
+		printf("Corrupting sector %llu (WRITE_LONG): ", lba);
+	}
+	fflush(stdout);
+
+	/* Try and ensure that the system doesn't have our sector in cache */
+	flush_buffer_cache(fd);
+
+	if (do_taskfile_cmd(fd, r, timeout_12secs)) {
+		err = errno;
+		perror("FAILED");
+	} else {
+		printf("succeeded\n");
+	}
+	free(r);
+	return err;
+}
+
+static int do_write_sector (int fd, __u64 lba, const char *devname)
+{
+	int err = 0;
+	__u8 ata_op;
+	struct hdio_taskfile *r;
+
+	abort_if_not_full_device(fd, lba, devname);
+	r = malloc(sizeof(struct hdio_taskfile) + 512);
+	if (!r) {
+		err = errno;
+		perror("malloc()");
+		return err;
+	}
+	ata_op = (lba >> 28) ? ATA_OP_WRITE_PIO_EXT : ATA_OP_WRITE_PIO;
+	init_hdio_taskfile(r, ata_op, RW_WRITE, LBA28_OK, lba, 1, 512);
+
+	printf("re-writing sector %llu: ", lba);
+	fflush(stdout);
+
+	if (do_taskfile_cmd(fd, r, timeout_12secs)) {
+		err = errno;
+		perror("FAILED");
+	} else {
+		printf("succeeded\n");
+	}
+
+	// Try and ensure that the system doesn't have our sector in cache:
+	flush_buffer_cache(fd);
+
+	free(r);
+	return err;
+}
+
+static int do_read_sector (int fd, __u64 lba, const char *devname)
+{
+	int err = 0;
+	__u8 ata_op;
+	struct hdio_taskfile *r;
+
+	abort_if_not_full_device(fd, lba, devname);
+	r = malloc(sizeof(struct hdio_taskfile) + 512);
+	if (!r) {
+		err = errno;
+		perror("malloc()");
+		return err;
+	}
+	ata_op = (lba >> 28) ? ATA_OP_READ_PIO_EXT : ATA_OP_READ_PIO;
+	init_hdio_taskfile(r, ata_op, RW_READ, LBA28_OK, lba, 1, 512);
+
+	printf("reading sector %llu: ", lba);
+	fflush(stdout);
+
+	if (do_taskfile_cmd(fd, r, timeout_12secs)) {
+		err = errno;
+		perror("FAILED");
+	} else {
+		printf("succeeded\n");
+		dump_sector(r->data);
+	}
+	free(r);
+	return err;
+}
+
+static int do_set_max_sectors (int fd, __u16 *id, __u64 max_lba, int permanent)
+{
+	int err = 0;
+	struct hdio_taskfile r;
+	__u8 nsect = permanent ? 1 : 0;
+
+	if ((max_lba >> 28) || (id && ((id[83] & 0xc400) == 0x4400) && (id[86] & 0x0400))) {
+		init_hdio_taskfile(&r, ATA_OP_SET_MAX_EXT, RW_READ, LBA48_FORCE, max_lba, nsect, 0);
+	} else {
+		init_hdio_taskfile(&r, ATA_OP_SET_MAX, RW_READ, LBA28_OK, max_lba, nsect, 0);
+		r.oflags.b.feat = 1;  /* this ATA op requires feat==0 */
+	}
+
+	/* spec requires that we do this immediately in front.. racey */
+	if (!do_get_native_max_sectors(fd, id))
+		return errno;
+
+	/* now set the new value */
+	if (do_taskfile_cmd(fd, &r, timeout_12secs)) {
+		err = errno;
+		perror(" SET_MAX_ADDRESS failed");
+	}
+	return err;
 }
 
 void process_dev (char *devname)
 {
 	int fd;
+	int err = 0;
 	static long parm, multcount;
 	__u16 *id = (void *)-1;
 
 	fd = open (devname, open_flags);
 	if (fd < 0) {
-		int err = errno;
+		if (errno == EROFS) {
+			open_flags &= ~O_WRONLY;
+			fd = open (devname, open_flags);
+			if (fd >= 0)
+				goto open_ok;
+		}
+		err = errno;
 		perror(devname);
 		exit(err);
 	}
+open_ok:
 	if (!quiet)
 		printf("\n%s:\n", devname);
 
 	if (set_fsreadahead) {
 		if (get_fsreadahead)
 			printf(" setting fs readahead to %d\n", fsreadahead);
-		if (ioctl(fd, BLKRASET, fsreadahead))
+		if (ioctl(fd, BLKRASET, fsreadahead)) {
+			err = errno;
 			perror(" BLKRASET failed");
+		}
 	}
 	if (set_unregister) {
 		if (get_unregister)
-			printf(" attempting to unregister hwif#%u\n", hwif);
-		if (ioctl(fd, HDIO_UNREGISTER_HWIF, hwif))
+			printf(" attempting to unregister hwif#%u\n", set_unregister);
+		if (ioctl(fd, HDIO_UNREGISTER_HWIF, set_unregister)) {
+			err = errno;
 			perror(" HDIO_UNREGISTER_HWIF failed");
+		}
 	}
 	if (scan_hwif) {
 		int	args[3];
@@ -834,8 +1120,10 @@ void process_dev (char *devname)
 		args[0] = hwif_data;
 		args[1] = hwif_ctrl;
 		args[2] = hwif_irq;
-		if (ioctl(fd, HDIO_SCAN_HWIF, args))
+		if (ioctl(fd, HDIO_SCAN_HWIF, args)) {
+			err = errno;
 			perror(" HDIO_SCAN_HWIF failed");
+		}
 	}
 	if (set_piomode) {
 		if (get_piomode) {
@@ -848,68 +1136,86 @@ void process_dev (char *devname)
 			else
 				printf(" attempting to set UDMA mode to %d\n", (piomode-200));
 		}
-		if (ioctl(fd, HDIO_SET_PIO_MODE, piomode))
+		if (ioctl(fd, HDIO_SET_PIO_MODE, piomode)) {
+			err = errno;
 			perror(" HDIO_SET_PIO_MODE failed");
+		}
 	}
 	if (set_io32bit) {
 		if (get_io32bit)
 			printf(" setting 32-bit IO_support flag to %d\n", io32bit);
-		if (ioctl(fd, HDIO_SET_32BIT, io32bit))
+		if (ioctl(fd, HDIO_SET_32BIT, io32bit)) {
+			err = errno;
 			perror(" HDIO_SET_32BIT failed");
+		}
 	}
 	if (set_mult) {
 		if (get_mult)
 			printf(" setting multcount to %d\n", mult);
-		if (ioctl(fd, HDIO_SET_MULTCOUNT, mult))
+		if (ioctl(fd, HDIO_SET_MULTCOUNT, mult)) {
+			err = errno;
 			perror(" HDIO_SET_MULTCOUNT failed");
+		}
 	}
 	if (set_readonly) {
 		if (get_readonly) {
 			printf(" setting readonly to %d", readonly);
 			on_off(readonly);
 		}
-		if (ioctl(fd, BLKROSET, &readonly))
+		if (ioctl(fd, BLKROSET, &readonly)) {
+			err = errno;
 			perror(" BLKROSET failed");
+		}
 	}
 	if (set_unmask) {
 		if (get_unmask) {
 			printf(" setting unmaskirq to %d", unmask);
 			on_off(unmask);
 		}
-		if (ioctl(fd, HDIO_SET_UNMASKINTR, unmask))
+		if (ioctl(fd, HDIO_SET_UNMASKINTR, unmask)) {
+			err = errno;
 			perror(" HDIO_SET_UNMASKINTR failed");
+		}
 	}
 	if (set_dma) {
 		if (get_dma) {
 			printf(" setting using_dma to %d", dma);
 			on_off(dma);
 		}
-		if (ioctl(fd, HDIO_SET_DMA, dma))
+		if (ioctl(fd, HDIO_SET_DMA, dma)) {
+			err = errno;
 			perror(" HDIO_SET_DMA failed");
+		}
 	}
 	if (set_dma_q) {
 		if (get_dma_q) {
 			printf(" setting DMA queue_depth to %d", dma_q);
 			on_off(dma_q);
 		}
-		if (ioctl(fd, HDIO_SET_QDMA, dma_q))
+		if (ioctl(fd, HDIO_SET_QDMA, dma_q)) {
+			err = errno;
 			perror(" HDIO_SET_QDMA failed");
+		}
 	}
 	if (set_nowerr) {
 		if (get_nowerr) {
 			printf(" setting nowerr to %d", nowerr);
 			on_off(nowerr);
 		}
-		if (ioctl(fd, HDIO_SET_NOWERR, nowerr))
+		if (ioctl(fd, HDIO_SET_NOWERR, nowerr)) {
+			err = errno;
 			perror(" HDIO_SET_NOWERR failed");
+		}
 	}
 	if (set_keep) {
 		if (get_keep) {
 			printf(" setting keep_settings to %d", keep);
 			on_off(keep);
 		}
-		if (ioctl(fd, HDIO_SET_KEEPSETTINGS, keep))
+		if (ioctl(fd, HDIO_SET_KEEPSETTINGS, keep)) {
+			err = errno;
 			perror(" HDIO_SET_KEEPSETTINGS failed");
+		}
 	}
 	if (set_doorlock) {
 		__u8 args[4] = {0,0,0,0};
@@ -918,8 +1224,10 @@ void process_dev (char *devname)
 			printf(" setting drive doorlock to %d", doorlock);
 			on_off(doorlock);
 		}
-		if (do_drive_cmd(fd, args))
+		if (do_drive_cmd(fd, args)) {
+			err = errno;
 			perror(" HDIO_DRIVE_CMD(doorlock) failed");
+		}
 	}
 	if (set_dkeep) {
 		/* lock/unlock the drive's "feature" settings */
@@ -929,24 +1237,30 @@ void process_dev (char *devname)
 			on_off(dkeep);
 		}
 		args[2] = dkeep ? 0x66 : 0xcc;
-		if (do_drive_cmd(fd, args))
+		if (do_drive_cmd(fd, args)) {
+			err = errno;
 			perror(" HDIO_DRIVE_CMD(keepsettings) failed");
+		}
 	}
 	if (set_defects) {
 		__u8 args[4] = {ATA_OP_SETFEATURES,0,0x04,0};
 		args[2] = defects ? 0x04 : 0x84;
 		if (get_defects)
 			printf(" setting drive defect management to %d\n", defects);
-		if (do_drive_cmd(fd, args))
+		if (do_drive_cmd(fd, args)) {
+			err = errno;
 			perror(" HDIO_DRIVE_CMD(defectmgmt) failed");
+		}
 	}
 	if (set_prefetch) {
 		__u8 args[4] = {ATA_OP_SETFEATURES,0,0xab,0};
 		args[1] = prefetch;
 		if (get_prefetch)
 			printf(" setting drive prefetch to %d\n", prefetch);
-		if (do_drive_cmd(fd, args))
+		if (do_drive_cmd(fd, args)) {
+			err = errno;
 			perror(" HDIO_DRIVE_CMD(setprefetch) failed");
+		}
 	}
 	if (set_xfermode) {
 		__u8 args[4] = {ATA_OP_SETFEATURES,0,3,0};
@@ -955,8 +1269,10 @@ void process_dev (char *devname)
 			printf(" setting xfermode to %d", xfermode_requested);
 			interpret_xfermode(xfermode_requested);
 		}
-		if (do_drive_cmd(fd, args))
+		if (do_drive_cmd(fd, args)) {
+			err = errno;
 			perror(" HDIO_DRIVE_CMD(setxfermode) failed");
+		}
 	}
 	if (set_lookahead) {
 		__u8 args[4] = {ATA_OP_SETFEATURES,0,0,0};
@@ -965,8 +1281,10 @@ void process_dev (char *devname)
 			printf(" setting drive read-lookahead to %d", lookahead);
 			on_off(lookahead);
 		}
-		if (do_drive_cmd(fd, args))
+		if (do_drive_cmd(fd, args)) {
+			err = errno;
 			perror(" HDIO_DRIVE_CMD(setreadahead) failed");
+		}
 	}
 	if (set_powerup_in_standby) {
 		__u8 args[4] = {ATA_OP_SETFEATURES,0,0,0};
@@ -986,8 +1304,10 @@ void process_dev (char *devname)
 		}
 		args[0] = ATA_OP_SETFEATURES;
 		args[2] = powerup_in_standby ? 0x06 : 0x86;
-		if (do_drive_cmd(fd, args))
+		if (do_drive_cmd(fd, args)) {
+			err = errno;
 			perror(" HDIO_DRIVE_CMD(powerup_in_standby) failed");
+		}
 	}
 	if (set_apmmode) {
 		__u8 args[4] = {ATA_OP_SETFEATURES,0,0,0};
@@ -1004,14 +1324,23 @@ void process_dev (char *devname)
 			if (get_apmmode)
 				printf(" 0x%02x (%d)\n",apmmode,apmmode);
 		}
-		if (do_drive_cmd(fd, args))
+		if (do_drive_cmd(fd, args)) {
+			err = errno;
 			perror(" HDIO_DRIVE_CMD failed");
+		}
 	}
 	if (set_cdromspeed) {
+		/* FIXME: the CDROM_SELECT_SPEED ioctl
+		 * actually issues GPCMD_SET_SPEED to the drive.
+		 * But many newer DVD drives want GPCMD_SET_STREAMING instead.
+		 * Perhaps we should just do an sg16 issue here of *both* opcodes?
+		 */
 		if (get_cdromspeed)
 			printf ("setting cdrom speed to %d\n", cdromspeed);
-		if (ioctl (fd, CDROM_SELECT_SPEED, cdromspeed))
+		if (ioctl (fd, CDROM_SELECT_SPEED, cdromspeed)) {
+			err = errno;
 			perror(" CDROM_SELECT_SPEED failed");
+		}
 	}
 	if (set_acoustic) {
 		__u8 args[4];
@@ -1021,8 +1350,10 @@ void process_dev (char *devname)
 		args[1] = acoustic;
 		args[2] = acoustic ? 0x42 : 0xc2;
 		args[3] = 0;
-		if (do_drive_cmd(fd, args))
+		if (do_drive_cmd(fd, args)) {
+			err = errno;
 			perror(" HDIO_DRIVE_CMD:ACOUSTIC failed");
+		}
 	}
 	if (set_wcache) {
 		if (get_wcache) {
@@ -1030,15 +1361,17 @@ void process_dev (char *devname)
 			on_off(wcache);
 		}
 		if (!wcache)
-			flush_wcache(fd, &id);
+			err = flush_wcache(fd, &id);
 		if (ioctl(fd, HDIO_SET_WCACHE, wcache)) {
 			__u8 setcache[4] = {ATA_OP_SETFEATURES,0,0,0};
 			setcache[2] = wcache ? 0x02 : 0x82;
-			if (do_drive_cmd(fd, setcache))
+			if (do_drive_cmd(fd, setcache)) {
+				err = errno;
 				perror(" HDIO_DRIVE_CMD(setcache) failed");
+			}
 		}
 		if (!wcache)
-			flush_wcache(fd, &id);
+			err = flush_wcache(fd, &id);
 	}
 	if (set_standbynow) {
 		__u8 args1[4] = {ATA_OP_STANDBYNOW1,0,0,0};
@@ -1046,8 +1379,10 @@ void process_dev (char *devname)
 		if (get_standbynow)
 			printf(" issuing standby command\n");
 		if (do_drive_cmd(fd, args1)
-		 && do_drive_cmd(fd, args2))
+		 && do_drive_cmd(fd, args2)) {
+			err = errno;
 			perror(" HDIO_DRIVE_CMD(standby) failed");
+		}
 	}
 	if (set_sleepnow) {
 		__u8 args1[4] = {ATA_OP_SLEEPNOW1,0,0,0};
@@ -1055,8 +1390,10 @@ void process_dev (char *devname)
 		if (get_sleepnow)
 			printf(" issuing sleep command\n");
 		if (do_drive_cmd(fd, args1)
-		 && do_drive_cmd(fd, args2))
+		 && do_drive_cmd(fd, args2)) {
+			err = errno;
 			perror(" HDIO_DRIVE_CMD(sleep) failed");
+		}
 	}
 	if (set_security) {
 		do_set_security(fd);
@@ -1064,15 +1401,19 @@ void process_dev (char *devname)
 	if (set_freeze) {
 		__u8 args[4] = {ATA_OP_SECURITY_FREEZE_LOCK,0,0,0};
 		printf(" issuing Security Freeze command\n");
-		if (do_drive_cmd(fd, args))
+		if (do_drive_cmd(fd, args)) {
+			err = errno;
 			perror(" HDIO_DRIVE_CMD(freeze) failed");
+		}
 	}
 	if (set_seagate) {
 		__u8 args[4] = {0xfb,0,0,0};
 		if (get_seagate)
 			printf(" disabling Seagate auto powersaving mode\n");
-		if (do_drive_cmd(fd, args))
+		if (do_drive_cmd(fd, args)) {
+			err = errno;
 			perror(" HDIO_DRIVE_CMD(seagatepwrsave) failed");
+		}
 	}
 	if (set_standby) {
 		__u8 args[4] = {ATA_OP_SETIDLE1,standby,0,0};
@@ -1080,33 +1421,61 @@ void process_dev (char *devname)
 			printf(" setting standby to %u", standby);
 			interpret_standby();
 		}
-		if (do_drive_cmd(fd, args))
+		if (do_drive_cmd(fd, args)) {
+			err = errno;
 			perror(" HDIO_DRIVE_CMD(setidle1) failed");
+		}
 	}
 	if (set_busstate) {
 		if (get_busstate)
 			printf(" setting bus state to %d (%s)\n", busstate, busstate_str(busstate));
-		if (ioctl(fd, HDIO_SET_BUSSTATE, busstate))
+		if (ioctl(fd, HDIO_SET_BUSSTATE, busstate)) {
+			err = errno;
 			perror(" HDIO_SET_BUSSTATE failed");
+		}
 	}
-	if (do_drq_hsm_error) {
+	if (set_max_sectors) {
+		if (get_native_max_sectors)
+			printf(" setting max visible sectors to %llu (%s)\n", set_max_addr, set_max_permanent ? "permanent" : "temporary");
+		id = get_identify_data(fd, id);
+		if (set_max_addr < get_lba_capacity(id))
+			confirm_i_know_what_i_am_doing("-Nnnnnn", "You have requested reducing the apparent size of the drive.\nThis is a BAD idea, and can easily destroy all of the drive's contents.");
+		err = do_set_max_sectors(fd, id, set_max_addr - 1, set_max_permanent);
+		id = (void *)-1; /* invalidate existing identify data */
+	}
+	if (make_bad_sector) {
+		id = get_identify_data(fd, id);
+		confirm_i_know_what_i_am_doing("--make-bad-sector", "You are trying to deliberately corrupt a low-level sector on the media\nThis is a BAD idea, and can easily result in total data loss.");
+		err = do_make_bad_sector(fd, id, make_bad_sector_addr, devname);
+	}
+	if (write_sector) {
+		confirm_i_know_what_i_am_doing("--write-sector", "You are trying to deliberately overwrite a low-level sector on the media\nThis is a BAD idea, and can easily result in total data loss.");
+		err = do_write_sector(fd, write_sector_addr, devname);
+	}
+	if (read_sector) {
+		err = do_read_sector(fd, read_sector_addr, devname);
+	}
+	if (drq_hsm_error) {
 		id = get_identify_data(fd, id);
 		if (id) {
 			__u8 args[4] = {0,0,0,0};
 			args[0] = last_identify_op;
-			sync();
 			printf(" triggering \"stuck DRQ\" host state machine error\n");
-			sync();
+			flush_buffer_cache(fd);
 			sleep(1);
 			do_drive_cmd(fd, args);
-			perror("do_drq_hsm_error");
+			err = errno;
+			perror("drq_hsm_error");
 			fprintf(stderr, "ata status=0x%02x ata error=0x%02x\n", args[0], args[1]);
 		}
 	}
+	id = (void *)-1; /* force re-IDENTIFY in case something above modified settings */
 	if (get_hitachi_temp) {
 		__u8 args[4] = {0xf0,0,0x01,0}; /* "Sense Condition", vendor-specific */
-		if (do_drive_cmd(fd, args))
+		if (do_drive_cmd(fd, args)) {
+			err = errno;
 			perror(" HDIO_DRIVE_CMD(hitachisensecondition) failed");
+		}
 		else {
 			printf(" drive temperature (celsius) is:  ");
 			if (args[2]==0)
@@ -1132,7 +1501,7 @@ void process_dev (char *devname)
 		if (0 == ioctl(fd, HDIO_GET_32BIT, &parm)) {
 			printf(" IO_support    =%3ld (", parm);
 			switch (parm) {
-				case 0:	printf("default ");
+				case 0:	printf("default) \n");
 				case 2: printf("16-bit)\n");
 					break;
 				case 1:	printf("32-bit)\n");
@@ -1143,12 +1512,18 @@ void process_dev (char *devname)
 					break;
 				default:printf("\?\?\?)\n");
 			}
+               } else {
+                       err = errno;
+                       perror(" HDIO_GET_32BIT failed");
 		}
 	}
 	if (do_defaults || get_unmask) {
 		if (0 == ioctl(fd, HDIO_GET_UNMASKINTR, &parm)) {
 			printf(" unmaskirq     = %2ld", parm);
 			on_off(parm);
+               } else {
+                       err = errno;
+                       perror(" HDIO_GET_UNMASKINTR failed");
 		}
 	}
 
@@ -1159,10 +1534,14 @@ void process_dev (char *devname)
 				printf(" (DMA-Assisted-PIO)\n");
 			else
 				on_off(parm);
+                } else {
+                       err = errno;
+                       perror(" HDIO_GET_DMA failed");
 		}
 	}
 	if (get_dma_q) {
 		if(ioctl(fd, HDIO_GET_QDMA, &parm)) {
+			err = errno;
 			perror(" HDIO_GET_QDMA failed");
 		} else {
 			printf(" queue_depth   = %2ld", parm);
@@ -1173,28 +1552,34 @@ void process_dev (char *devname)
 		if (0 == ioctl(fd, HDIO_GET_KEEPSETTINGS, &parm)) {
 			printf(" keepsettings  = %2ld", parm);
 			on_off(parm);
+		} else {
+			err = errno;
+                        perror(" HDIO_GET_KEEPSETTINGS failed");
 		}
 	}
 	if (get_nowerr) {
-		if (ioctl(fd, HDIO_GET_NOWERR, &parm))
+		if (ioctl(fd, HDIO_GET_NOWERR, &parm)) {
+			err = errno;
 			perror(" HDIO_GET_NOWERR failed");
-		else {
+		} else {
 			printf(" nowerr        = %2ld", parm);
 			on_off(parm);
 		}
 	}
 	if (do_defaults || get_readonly) {
-		if (ioctl(fd, BLKROGET, &parm))
+		if (ioctl(fd, BLKROGET, &parm)) {
+			err = errno;
 			perror(" BLKROGET failed");
-		else {
+		} else {
 			printf(" readonly      = %2ld", parm);
 			on_off(parm);
 		}
 	}
 	if (do_defaults || get_fsreadahead) {
-		if (ioctl(fd, BLKRAGET, &parm))
+		if (ioctl(fd, BLKRAGET, &parm)) {
+			err = errno;
 			perror(" BLKRAGET failed");
-		else {
+		} else {
 			printf(" readahead     = %2ld", parm);
 			on_off(parm);
 		}
@@ -1206,15 +1591,17 @@ void process_dev (char *devname)
  * Note to self:  when compiled 32-bit (AMD,Mips64), the userspace version of this struct
  * is going to be 32-bits smaller than the kernel representation.. random stack corruption!
  */
-		static struct hd_geometry g;
+		static struct local_hd_geometry      g;
 		static struct local_hd_big_geometry bg;
 
-		if (0 == do_blkgetsize(fd, &blksize)) {
+		err = do_blkgetsize(fd, &blksize);
+		if (!err) {
 			if (!ioctl(fd, HDIO_GETGEO_BIG, &bg))
 				printf(msg, bg.cylinders, bg.heads, bg.sectors, blksize, bg.start);
-			else if (ioctl(fd, HDIO_GETGEO, &g))
+			else if (ioctl(fd, HDIO_GETGEO, &g)) {
+				err = errno;
 				perror(" HDIO_GETGEO failed");
-			else
+			} else
 				printf(msg, g.cylinders, g.heads, g.sectors, blksize, g.start);
 		}
 	}
@@ -1223,9 +1610,10 @@ void process_dev (char *devname)
 		const char *state;
 		if (do_drive_cmd(fd, args)
 		 && (args[0] = ATA_OP_CHECKPOWERMODE2) /* (single =) try again with 0x98 */
-		 && do_drive_cmd(fd, args))
+		 && do_drive_cmd(fd, args)) {
+			err = errno;
 			state = "unknown";
-		else
+		} else
 			state = (args[2] == 255) ? "active/idle" : "standby";
 		printf(" drive state is:  %s\n", state);
 	}
@@ -1240,21 +1628,18 @@ void process_dev (char *devname)
 			dump_identity(id2);
 		} else if (errno == -ENOMSG)
 			printf(" no identification info available\n");
-		else
+		else {
+			err = errno;
 			perror(" HDIO_GET_IDENTITY failed");
+		}
 	}
 	if (do_IDentity) {
 		id = get_identify_data(fd, id);
 		if (id) {
-			int i;
-			if (do_IDentity == 2) {
-				for (i = 0; i < (256/8); ++i) {
-					printf("%04x %04x %04x %04x %04x %04x %04x %04x\n", id[0], id[1], id[2], id[3], id[4], id[5], id[6], id[7]);
-					id += 8;
-				}
-			} else {
+			if (do_IDentity == 2)
+				dump_sector(id);
+			else
 				identify((void *)id);
-			}
 		}
 	}
 	if (get_lookahead) {
@@ -1294,41 +1679,68 @@ void process_dev (char *devname)
 		}
 	}
 	if (get_busstate) {
-		if (ioctl(fd, HDIO_GET_BUSSTATE, &parm))
+		if (ioctl(fd, HDIO_GET_BUSSTATE, &parm)) {
+			err = errno;
 			perror(" HDIO_GET_BUSSTATE failed");
-		else
+		} else
 			printf(" busstate      = %2ld (%s)\n", parm, busstate_str(parm));
 	}
+	if (get_native_max_sectors) {
+		__u64 visible, native;
+		id = get_identify_data(fd, id);
+		if (id) {
+			visible = get_lba_capacity(id);
+			native  = do_get_native_max_sectors(fd, id);
+			if (!native) {
+				err = errno;
+			} else {
+				printf(" max sectors   = %llu/%llu", visible, native);
+				if (visible < native)
+					printf(", HPA is enabled\n");
+				else if (visible == native)
+					printf(", HPA is disabled\n");
+				else
+					printf(", HPA setting seems invalid\n");
+			}
+		}
+	}
 
 	if (do_ctimings)
 		time_cache(fd);
 	if (do_flush_wcache)
-		flush_wcache(fd, &id);
+		err = flush_wcache(fd, &id);
 	if (do_timings)
-		time_device(fd);
+		err = time_device(fd);
 	if (do_flush)
 		flush_buffer_cache(fd);
 	if (set_reread_partn) {
 		if (get_reread_partn)
 			printf(" re-reading partition table\n");
 		if (ioctl(fd, BLKRRPART, NULL)) {
+			err = errno;
 			perror(" BLKRRPART failed");
 		}
 	}
 	if (set_doreset) {
 		if (get_doreset)
 			printf(" resetting drive\n");
-		if (ioctl(fd, HDIO_DRIVE_RESET, NULL))
+		if (ioctl(fd, HDIO_DRIVE_RESET, NULL)) {
+			err = errno;
 			perror(" HDIO_DRIVE_RESET failed");
+		}
 	}
 	if (set_tristate) {
 		__u8 args[4] = {0,tristate,0,0};
 		if (get_tristate)
 			printf(" setting tri-state to %d\n", tristate);
-		if (ioctl(fd, HDIO_TRISTATE_HWIF, &args))
+		if (ioctl(fd, HDIO_TRISTATE_HWIF, &args)) {
+			err = errno;
 			perror(" HDIO_TRISTATE_HWIF failed");
+		}
 	}
 	close (fd);
+	if (err)
+		exit (err);
 }
 
 static void usage_help (int rc)
@@ -1359,6 +1771,7 @@ static void usage_help (int rc)
 	" -L   set drive doorlock (0/1) (removable harddisks only)\n"
 	" -M   get/set acoustic management (0-254, 128: quiet, 254: fast)\n"
 	" -m   get/set multiple sector count\n"
+	" -N   get/set max visible number of sectors (HPA) (VERY DANGEROUS)\n"
 	" -n   get/set ignore-write-errors flag (0/1)\n"
 	" -p   set PIO mode on IDE interface chipset (0,1,2,3,4,...)\n"
 	" -P   set drive prefetch count\n"
@@ -1382,12 +1795,15 @@ static void usage_help (int rc)
 	" -Y   put drive to sleep\n"
 	" -Z   disable Seagate auto-powersaving mode\n"
 	" -z   re-read partition table\n"
-	" --direct         use O_DIRECT to bypass page cache for timings\n"
-	" --Istdin         read identify data from stdin as ASCII hex\n"
-	" --Istdout        write identify data to stdout as ASCII hex\n"
-	" --verbose        display extra diagnostics from some commands\n"
-	" --security-help  display help for ATA security commands\n"
-	" --drq-hsm-error  crash system with a \"stuck DRQ\" error (VERY DANGEROUS)\n"
+	" --direct          use O_DIRECT to bypass page cache for timings\n"
+	" --drq-hsm-error   crash system with a \"stuck DRQ\" error (VERY DANGEROUS)\n"
+	" --Istdin          read identify data from stdin as ASCII hex\n"
+	" --Istdout         write identify data to stdout as ASCII hex\n"
+	" --make-bad-sector deliberately corrupt a sector directly on the media (VERY DANGEROUS)\n"
+	" --read-sector     read and dump (in hex) a sector directly from the media\n"
+	" --security-help   display help for ATA security commands\n"
+	" --verbose         display extra diagnostics from some commands\n"
+	" --write-sector    repair/overwrite a (possibly bad) sector directly on the media (VERY DANGEROUS)\n"
 	"\n");
 	exit(rc);
 }
@@ -1441,7 +1857,7 @@ static void security_help (int rc)
 			flag = 1;				\
 	} while (0)
 
-static int fromhex (unsigned char c)
+static int fromhex (__u8 c)
 {
 	if (c >= '0' && c <= '9')
 		return (c - '0');
@@ -1553,6 +1969,52 @@ static void get_security_password (int handle_NULL)
 		++argp;
 }
 
+static int
+get_sector_value (int optional, const char flag_c, int *flag_p, __u64 *value_p, unsigned int min_value, const char *emsg)
+{
+	int got_value = 0, got_digit = 0;
+	const __u64 limit = (1ULL << 48) - 1;
+	__u64 value = *value_p;
+
+	if (!optional)
+		got_value = 1;	/* makes getting a value mandatory here */
+
+	if (!*argp && argc && (isdigit(**argv) || (flag_p && flag_c == **argv)))
+		argp = *argv++, --argc;
+
+	if (flag_p) {
+		*flag_p = 0;
+		if (*argp == flag_c) {
+			*flag_p = 1;
+			got_value = 1;
+			argp++;
+		}
+	}
+
+	while (isdigit(*argp)) {
+		if (!got_digit++)
+			value = 0;
+		got_value = 1;
+		value = (value * 10) + (*argp++ - '0');
+	}
+	if (got_value) {
+		if (value < min_value || value > limit) {
+			fprintf(stderr, "  %s: bad/missing sector value\n", emsg);
+			exit(EINVAL);
+		}
+		*value_p = value;
+	}
+	return got_value;
+}
+
+static void
+get_set_max_sectors_parms (void)
+{
+	get_native_max_sectors = noisy;
+	noisy = 1;
+	set_max_sectors = get_sector_value(1, 'p', &set_max_permanent, &set_max_addr, 1, "-N");
+}
+
 static void
 handle_standalone_longarg (char *name)
 {
@@ -1617,7 +2079,16 @@ get_longarg (void)
 		open_flags |= O_DIRECT;
 		--num_flags_processed;	/* doesn't count as an action flag */
 	} else if (0 == strcasecmp(name, "drq-hsm-error")) {
-		do_drq_hsm_error = 1;
+		drq_hsm_error = 1;
+	} else if (0 == strcasecmp(name, "make-bad-sector")) {
+		make_bad_sector = 1;
+		get_sector_value(0, 'f', &make_bad_sector_flagged, &make_bad_sector_addr, 0, "--make-bad-sector");
+	} else if (0 == strcasecmp(name, "write-sector") || 0 == strcasecmp(name, "repair-sector")) {
+		write_sector = 1;
+		get_sector_value(0, 0, NULL, &write_sector_addr, 0, "--write-sector");
+	} else if (0 == strcasecmp(name, "read-sector")) {
+		read_sector = 1;
+		get_sector_value(0, 0, NULL, &read_sector_addr, 0, "--read-sector");
 	} else if (0 == strcasecmp(name, "Istdout")) {
 		do_IDentity = 2;
 	} else if (0 == strcasecmp(name, "security-mode")) {
@@ -1711,6 +2182,7 @@ int main (int _argc, char **_argv)
 				case GET_SET_PARM('m',"multmode-count",mult,0,64);
 				case GET_SET_PARM('M',"acoustic-management",acoustic,0,255);
 				case GET_SET_PARM('n',"ignore-write-errors",nowerr,0,1);
+				case              'N': get_set_max_sectors_parms(); break;
 				case     SET_PARM('P',"prefetch",prefetch,0,255);
 				case              'q': quiet = 1; noisy = 0; break;
 				case     SET_PARM('Q',"queue-depth",dma_q,0,32);
